"""EGDAR index downloader.

This module downloads relevant index files from the EGDAR database,
consolidates these. Moreover, filters are applied to only return relevant
results.

Functions in this module:
    download
    consolidates
    filter
"""

import urllib.request
import urllib.error
import os.path
import pandas as pd


class EdgarIndex(object):
    """Index of edgarfilings.

    Attributes:
        sample_start (datetime): Start date for the filings in sample.
        sample_end (datetime): End date for the filings in sample.
        dir_work (str): Path to working directory.
        sub_index (str): Path to subdirectory in working directory.
        edgar_url (str): URL to EDGAR archive parent folder on server.

    """

    def __init__(self, sample_start, sample_end,
                 dir_work, sub_index, edgar_url):
        """Class construnctor."""
        self.sample_start = sample_start
        self.sample_end = sample_end
        self.dir_work = dir_work
        self.sub_index = sub_index
        self.edgar_url = edgar_url

    def download(self):
        """Download index files from the EDGAR database.

        Args:
            None

        Returns:
            None

        Raises:
            None

        """
        url_start = "edgar/full-index/"
        url_end = "/form.idx"

        # Exclude quarters which are not in sample
        start_excl = []
        if self.sample_start.month > 3:
            start_excl.append(1)
        if self.sample_start.month > 6:
            start_excl.append(2)
        if self.sample_start.month > 9:
            start_excl.append(3)

        end_excl = []
        if self.sample_end.month < 10:
            end_excl.append(4)
        if self.sample_end.month < 7:
            end_excl.append(3)
        if self.sample_end.month < 4:
            end_excl.append(2)

        # For each quarter in each year in sample, check whether quarter is in
        # sample. If so download the index file if it does not already exist
        for year in range(self.sample_start.year, self.sample_end.year + 1):
            for quarter in range(1, 5):
                if (year == self.sample_start.year) & (quarter in start_excl):
                    continue
                elif (year == self.sample_end.year) & (quarter in end_excl):
                    continue
                else:
                    url = (self.edgar_url
                           + url_start
                           + str(year)
                           + "/QTR"
                           + str(quarter)
                           + url_end
                           )
                    fname = (self.dir_work
                             + self.sub_index
                             + str(year)
                             + "_"
                             + str(quarter)
                             + ".txt"
                             )
                    if os.path.isdir(os.path.dirname(fname)) is False:
                        os.makedirs(os.path.dirname(fname))
                    if os.path.isfile(fname) is False:
                        try:
                            urllib.request.urlretrieve(url, fname)
                        except urllib.error.HTTPError as e:
                            if e.code == 404:
                                print(("%s not on server!", url))
                            else:
                                print(e)

    def consolidate(self):
        """Consolidate index files and filter for desired form type.

        Args:
            None

        Returns:
            None

        Raises:
            None

        """
        index_files = os.listdir(self.dir_work + self.sub_index)
        result = pd.DataFrame()

        if len(index_files) == 0:
            print("No index files found. Download index files first")
            return None

        for indexf in index_files:
            indexdf = pd.read_fwf(self.dir_work + self.sub_index + indexf,
                                  widths=[12, 62, 12, 12, 52],
                                  skiprows=8,
                                  comment="---")
            indexdf["date"] = pd.to_datetime(indexdf["Date Filed"])
            result = result.append(indexdf)

        self.cons_index = result

    def filter(self, sample_start, sample_end, filter_formtype, filter_CIK):
        """Filter the index based on date, form type and/or CIK.

        Args:
            sample_start (datetime): Start date for the filings in sample.
            sample_end (datetime): End date for the filings in sample.
            filter_formtype (list of str, optional): SEC form types to include.
                None deactives the filter. Defaults to none.
            filter_CIK (list of str, optional): Company CIKs to include.
                None deactives the filter. Defaults to None.

        Returns:
            None

        Raises:
            None

        """
        # Filter based on sample period
        filterdf = self.cons_index.loc[(self.cons_index.date >= sample_start) &
                                       (self.cons_index.date <= sample_end)]

        # Filter based on form type
        if filter_formtype is not None:
            filterdf = filterdf[filterdf["Form Type"].isin(filter_formtype)]

        # Filter based on CIK
        if filter_CIK is not None:
            filterdf = filterdf[filterdf["CIK"].isin(filter_CIK)]

        self.filtered_index = filterdf



"""Crawl EDGAR database to download index files and filings.

Example:
    import edgarsearch.edgarsearch
    search = edgarsearch.edgarsearch.search("20151001",
                                            "20161231",
                                            sample_size=200,
                                            filter_formtype=["8-K"])
    search.downloadindex()
    search.safe_download("months", 1, text_only=True,
                         fname_form="%Y/%m/%Y%m_%company",
                         chunk_size=100)

Todo:
    *  Add proper exception handling
    *  Add rollback functions if an error occurs
    *  Add testing

"""
import datetime
import edgarsearch.indexhandler as ih
import edgarsearch.filingsbatch as fb
import pandas as pd
from tqdm import tqdm


class Search(object):
    """Search of the EDGAR database.

    Attributes:
        sample_start (str):  Start date for filings in sample.
            The valid format is "YYYYMMDD".
        sample_end (str): End date for filings in sample.
            The valid format is "YYYYMMDD".
        sample_size: (int): Number of samples in scope. If -1, no limit is set.
            Default to -1.
        filter_formtype (list of str, optional): Filter based on filings type.
            Example:["8-K, 10-K"]. Defaults to None.
        filter_CIK (list of str, optional):Filter based on CIK.
            Example: ["12345678", "98765432"]. Defaults to None.
        dir_work (str): Working subdirectory for all saved data.
            Defaults to "edgar/".
        sub_index (str): Subdirectory in dir_work for saved index data.
            Defaults to "index/".
        sub_filings (str): Subdirectory in dir_work for saved filings data.
            Defaults to "filings".
        edgar_url (str): URL to EDGAR server
            Defaults to  "https://www.sec.gov/Archives/".

    """

    def __init__(self, sample_start, sample_end, sample_size=-1, *,
                 filter_formtype=None, filter_CIK=None, dir_work="edgar/",
                 sub_index="index/", sub_filings="filings/",
                 edgar_url="https://www.sec.gov/Archives/"):
        """Create new EDGAR search object."""
        self.dir_work = dir_work
        self.sub_index = sub_index
        self.sub_filings = sub_filings
        self.edgar_url = edgar_url
        try:
            sample_start = datetime.datetime.strptime(sample_start, "%Y%m%d")
            sample_end = datetime.datetime.strptime(sample_end, "%Y%m%d")
        except ValueError:
            print("Dates must be in YYYYMMDD format.")

        if sample_start > sample_end:
            raise ValueError("Sample start must be prior to sample end")
        self.sample_start = sample_start
        self.sample_end = sample_end
        self.filter_formtype = filter_formtype
        self.filter_CIK = filter_CIK
        self.sample_size = sample_size

    def downloadindex(self):
        """Download the index of the corresponding search.

        Args:
            None

        Returns:
            None

        Raises:
            None

        """
        e_index = ih.EdgarIndex(self.sample_start, self.sample_end,
                                self.dir_work, self.sub_index,
                                self.edgar_url)
        e_index.download()
        e_index.consolidate()
        self.full_index = e_index.cons_index
        e_index.filter(self.sample_start, self.sample_end,
                       self.filter_formtype,
                       self.filter_CIK)
        filt_index = e_index.filtered_index

        if self.sample_size > 0:
            self.cur_index = filt_index.sample(self.sample_size)
        else:
            self.cur_index = filt_index

    def downloadfilings(self, index=None, raw=False, text_only=True,
                        chunk_size=100, **kwargs):
        """Process download requests in chunks.

        The method will execute the following steps:
            1. Split urls into chunks
            For each chunk:
            2. Download the filings from the server and store the .txt files
               temporaryly
            If raw = False:
            3. Extract the containing files from the .txt files
            4. Fix broken image paths in documents
            5. Delete temporary .txt files
            6. Store information about stored documents

        Args:
            urls (list of str): URLs of EDGAR filings to process.
            index (pandas dataframe): Index dataframe to be downloaded.
                If none, self.cur_index will be used. Defaults to None.
            raw (bool): If true, the original .txt files from the EDGAR server
                will be stored. If false, the containing documents will be
                extracted and stored. Defaults to False.
            chunk_size (int): Number of filings to process in only iteration.
                The bigger the chunk, the bigger the temporary file cache.
                Defaults to 100.
        Keyword Args:
            text_only (bool): If True, only html and txt files are saved.
                All other (media) files are discarded. Defaults to True.
            fname_form (str): String with the filename format.
                Possible parameters:
                * %org - orignal filename on server
                * %company - Company Name
                * All date variables based on strftime.
                  See: http://strftime.org/
                Defaults to "%Y%m%d_%company_".

        Returns:
            None

        Raises:
            None

        """
        if index is None:
            df = self.cur_index
        else:
            df = index
        length = df.shape[0]
        # If text_only is false, media files are saved, which might require a
        # lotof disk space. Therefore, user is asked for confirmation
        if text_only is False:
            space_req = length / 1000
            string = ("Saving non-text files may take a lot of space."
                      "The estimated required space is %0.1f GB."
                      " Do you want to continue? [Y]es or [N]o? " % space_req)
            while True:
                check = input(string).lower()
                if ((check == "y") | (check == "yes") |
                   (check == "no") | (check == "n")):
                    break
            if (check == "no") | (check == "n"):
                return None, None

        # Check if chunk_size is reasonable and adjust if not
        if chunk_size > length:
            chunk_size = length

        # Create chunks and prepare empty output DataFrames
        c_list = [df.iloc[i:i + chunk_size, :]
                  for i in range(0, length, chunk_size)]
        final_tmp_f = pd.DataFrame()
        final_docs = pd.DataFrame()

        # Print information to user
        print("Total filings to download: %s" % length)
        print("Number of batches: %s (containg %s filings each)"
              % (len(c_list), chunk_size))
        print("Progress:")

        # Iterate over chunks, while showing a progess bar
        for index_chunk in tqdm(c_list):

            # Download the files from the server
            batch = fb.Batch(index_chunk, dir_work=self.dir_work,
                             sub_filings=self.sub_filings,
                             edgar_url=self.edgar_url)
            batch.download()

            # If raw is False, process the downloaded filins
            if raw is False:
                batch.splitfiles(**kwargs)
                documents = batch.docs
                batch.replace_img()
                batch.delete_tempfiles()
                final_docs = pd.concat([final_docs, documents])
            tmp_files = batch.temp_files
            final_tmp_f = pd.concat([final_tmp_f, tmp_files])

        self.temp_files = final_tmp_f
        self.docs = final_docs

    def safe_download(self, sm_type="num", sm_val=10000, **kwargs):
        """Run the data pipeline to download index and fillings data.

        Args:
            sm_type (str, optional): Defines the safe_mode to be used.
                Possible values:
                * "None": The full sample will be downloaded
                * "num" : Split sample based on the number of samples.
                * "years": Split sample based on sample years.
                * "months": Split sample based on sample months.
                Defaults to "num".
            sm_val (int, optional): Value for the safe mode.
                If safe_mode is "num": Number of samples per subsample.
                If safe_mode is "years": Number of years per subsample.
                If safe_mode is "months": Number of months per subsample.
                Defaults to 10000.
        Keyword arguments:
            raw (bool): If true, the original .txt files from the EDGAR server
                will be stored. If false, the containing documents will be
                extracted and stored. Defaults to False.
            text_only (bool): If True, only html and txt files are saved.
                All other (media) files are discarded. Defaults to True.
            chunk_size (int): Number of filings to process in only iteration.
                The bigger the chunk, the bigger the temporary file cache.
                Defaults to 100.
            fname_form (str): String with the filename format.
                Possible parameters:
                * %org - orignal filename on server
                * %company - Company Name
                * All date variables based on strftime.
                  See: http://strftime.org/
                Defaults to "%Y%m%d_%company_".
        Returns:
            None

        Raises:
            None

        """
        fulldf = self.cur_index

        if sm_type == "num":
            # Split sample into subsamples based on sample size
            for x in range(0, fulldf.shape[0], sm_val):
                limit = min(fulldf.shape[0], x + sm_val)
                tempdf = fulldf.iloc[x:limit, :]
                tqdm.write("Download sample %s-%s from total sample(size: %s)"
                           % (x, limit, fulldf.shape[0]))
                self.downloadfilings(index=tempdf, **kwargs)
                self.docs.to_csv("filings_"
                                 + str(x)
                                 + "_"
                                 + str(limit)
                                 + ".csv")
        elif sm_type == "years":
            # Split sample into subsamples based on years in sample period
            year_s = self.sample_start.year
            year_e = self.sample_end.year
            for period in range(year_s, year_e + 1, sm_val):
                end_year = min(year_e, period + sm_val - 1)
                tempdf = fulldf.loc[((fulldf.date >= str(period)) &
                                     (fulldf.date <= str(end_year + 1)))]
                tqdm.write("Download sample %s-%s from total sample %s-%s"
                           % (period, end_year, year_s, year_e))
                self.downloadfilings(index=tempdf, **kwargs)
                self.docs.to_csv("filings_"
                                 + str(period)
                                 + "_"
                                 + str(end_year)
                                 + ".csv")
        elif sm_type == "months":
            # Split sample into subsamples based on months in sample period
            starty = self.sample_start.year
            startm = self.sample_start.month
            endy = self.sample_end.year
            endm = self.sample_end.month
            startdate = str(starty) + str(startm).zfill(2)
            enddate = str(endy) + str(endy).zfill(2)
            for year in range(starty, endy + 1):
                if year != starty:
                    start = 1
                else:
                    start = startm

                if year != endy:
                    end = 13
                else:
                    end = endm + 1
                for month in range(start, end, sm_val):
                    limit = min(end, month + sm_val)
                    start = str(year) + "-" + str(month).zfill(2)
                    if limit < 13:
                        stop = str(year) + "-" + str(limit).zfill(2)
                    else:
                        stop = str(year + 1) + "-" + str(limit - 12).zfill(2)

                    limit_d = min(end, month + sm_val) - 1
                    if limit_d < 13:
                        stop_d = str(year) + "-" + str(limit_d).zfill(2)
                    else:
                        stop_d = (str(year + 1)
                                  + "-"
                                  + str(limit_d - 12).zfill(2)
                                  )
                    print("%s - %s" % (start, stop_d))

                    tempdf = fulldf.loc[((fulldf.date >= str(start)) &
                                         (fulldf.date < str(stop)))]
                    tqdm.write(("Download filings in months %s - %s from "
                               "total sample period %s-%s")
                               % (start, stop_d, startdate, enddate))
                    self.downloadfilings(index=tempdf, **kwargs)
                    self.docs.to_csv("filings_"
                                     + str(start)
                                     + "_"
                                     + str(stop_d)
                                     + ".csv")



"""Downloads filings from the EDGAR database.

Examples:
    simple:
    singledownload("edgar/data/1645148/0001213900-15-004775.txt")
    for slow internet connections:
    singledownload("edgar/data/1645148/0001213900-15-004775.txt", timeout=120)
    with different folder structure:
    singledownload("edgar/data/1645148/0001213900-15-004775.txt",
                   folder="myfolder/", sub="")

"""
import urllib.request
import uuid
import datetime
import os


def singledownload(url, edgar_url="https://www.sec.gov/Archives/",
                   folder="data/", sub="filings/", timeout=30):
    """Download filings from the EDGAR database.

    Args:
        url (str): Relative path on the EDGAR server.
        edgar_url (str): URL to EDGAR archive parent folder on server.
            Defaults to "https://www.sec.gov/Archives/".
        folder (str): Path to working directory. Defaults to "data/".
        sub (str): Path to subdirectory in working directory.
            Defaults to "filings/".
        timeout (int): Number of seconds to wait for the download to complete
            before the download attempt is counted as failed.
            Defaults to 30 seconds.

    Returns:
        result (list): Information on which was downloaded, the local filename
            and the date and time the file was downloaded.

    Raises:
        None

    """
    full_url = edgar_url + url
    fname = str(uuid.uuid4()) + ".txt"
    accessed = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
    try:
        txt = urllib.request.urlopen(full_url, timeout=timeout).read()
    except Exception as e:
        return [url, e, "Error"]

    if os.path.isdir(folder + sub) is False:
        try:
            os.makedirs(folder + sub)
        except Exception as e:
            print(e)

    try:
        f = open(folder + sub + fname, 'x')
    except:
        f = open(folder + sub + fname, 'w+')
    f.write(txt.decode('utf-8'))
    f.close
    result = [url, fname, accessed]
    return result



"""Useful tool collection.

Consists of the following functions:
    * getalpha: Convert integer to alphabetic character.
    * finduniquefilename: Find unique name if a file with the original filename
                          already exists.


Example:
    getalpha(0)  --> "A"
    getalpha(1)  --> "B"
    getalpha(27) --> "AB"
    finduniquefilename(ab, mode="alpha", exact=True):


Todo:
    *  None

"""

import os
import string
import glob


def getalpha(x):
    """Convert integer to alphabetic character.

    Assigns every integer to a alphabetic string.
    Examples for clarification:
        getalpha(0)  --> "A"
        getalpha(1)  --> "B"
        getalpha(27) --> "AB"

    Args:
        x (integer): Number to be converted

    Returns:
        result (String): Alphabetic equivalent to integer

    Raises:
        None

    """
    result = ""
    if x // 26 > 0:
        result += getalpha((x // 26)-1)
        result += string.ascii_uppercase[(x % 26)]
    else:
        result += string.ascii_uppercase[(x % 26)]
    return result


def finduniquefname(fname, mode="alpha", exact=True, **kwargs):
    """Find unique name if a file with the original file name already exists.

    Args:
        fname_form (String): Filename (incl. path) to be checked
        mode (String): Sets the mode for creating unique filenames.
            Possible values:
            * "alpha" will add alphabetical suffixes
            * "num"  will add numerical suffixes
            default("alpha")
        extract (Bool): If true, checks for the exact file name.
            If false, checks for the filename pattern
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        fname (String): Unique filename

    Raises:
        None

    """
    i = 0
    fpre, fsuf = os.path.splitext(fname)
    if exact is True:
        while os.path.isfile(fname):
            if mode == "alpha":
                fname = (
                         fpre
                         + getalpha(i)
                         + fsuf
                         )
            elif mode == "num":
                fname = (
                         fpre
                         + str(i)
                         + fsuf
                         )
            i += 1
    else:
        while len(glob.glob(fname + "*")) > 0:
            if mode == "alpha":
                fname = (
                         fpre
                         + getalpha(i)
                         + fsuf
                         )
            elif mode == "num":
                fname = (
                         fpre
                         + str(i)
                         + fsuf
                         )
            i += 1
    return(fname)



from setuptools import setup

setup(name="edgarsearch",
      version="0.2",
      description="Crawl EDGAR database to download index files and filings.",
      keywords="EDGAR index filings 8-K 10-K",
      url="https://github.com/markdembo/edgarsearch",
      download_url="https://github.com/markdembo/edgarsearch/archive/0.2.tar.gz",
      author="Mark Dembo",
      author_email="mark.dembo@student.unisg.ch",
      license="MIT",
      packages=["edgarsearch"],
      install_requires=[
          "pandas",
          "tqdm"
      ],
      )



"""

This a sample script to showcase the egdarsearch script.

For function defintion and possible parameters check out definition in
the edgarsearch.py file.

"""
import edgarsearch.edgarsearch

if __name__ == '__main__':
    # Setup the basical variables for the following commands (with defaults)
    my_edgar = edgarsearch.edgarsearch.edgar()
    # Define a search by passing the start and end of the sample period,
    # as well as the desired formtype

    my_edgar.definesearch("20150101", "20150731", filter_formtype=["8-K"])
    # Get the index file based on the defined search
    my_edgar.getindex()

    # Get a sample of 10 filings (including all media data) from the defined
    # search and store these using the desired file name pattern
    my_edgar.getfilings(text_only=False, sample_size=10,
                        fname_form="%Y/%m/%Y%m_%company")

    # Display the pandas df containg all the downloaded filings documents
    my_edgar.cur_filings



"""Filings downloader and cleaner.

This module downloads filings from the EDGAR database and splits the filings
into multiple files.

Example:
    batch = fb.Batch(indexdf, dir_work="edgar/",
                     sub_filings="filings/",
                     edgar_url="https://www.sec.gov/Archives/")
    batch.download()
    batch.splitfiles()
    batch.replace_img()
    batch.delete_tempfiles()
    batch.docs

"""
import pandas as pd
import multiprocessing as mp
import edgarsearch.mpworker as mpw
import edgarsearch.tools as t
import re
import uu
import os
import datetime
import itertools
import io
import time
from tqdm import tqdm


class Batch(object):
    """Batch of filings.

    Attributes:
        index_slice (pandas dataframe): From indexhandler
        show_progess (bool): If true, a progressbar on file level will be
            displayed. Defaults to True.
        dir_work: (str): Working directory for all saved data.
            Defaults to "edgar/".
        sub_filings (str): Subdirectory for filings data.
            Defaults to "sub_filings".
        edgar_url (str): URL to EDGAR server

    """

    def __init__(self, index_slice, show_progess=True,
                 dir_work="edgar/", sub_filings="filings/",
                 edgar_url="https://www.sec.gov/Archives/"):
        """Class construnctor."""
        self.index_slice = index_slice
        self.show_progress = show_progess
        self.dir_work = dir_work
        self.sub_filings = sub_filings
        self.edgar_url = edgar_url

        # List of str from filings index slice
        self.url_list = index_slice["File Name"].tolist()

    def collect(self, result):
        """Collect the results from the worker.

        Filter the results from the filings downloader and append to either the
        result list or error list.

        Args:
            result (list of str/int): Output from the singledownload worker.

        Returns:
            None

        Raises:
            None

        """
        global pbar2
        if result[2] != "Error":
            self.results.append(result)
            if self.show_progress:
                self.bar.update(1)
        else:
            self.errors.append(result)

    def download(self):
        """Download all filings from a list of urls.

        Returns:
            None

        Raises:
            None

        """
        # Create empty lists for errors and results
        self.errors = []
        self.results = []
        # Store the original URLs and set attemps = 0
        attempt = 0
        urls_queue = self.url_list
        # Init the second (lower) progress bar
        if self.show_progress:
            self.bar = tqdm(total=len(self.url_list))

        # If not all filins are downloaded yet and the attempts are below
        # three then try to download
        while len(self.results) < len(self.url_list) and attempt < 2:
            # Use multiprocessing pool to download files
            mplist = list(itertools.product(urls_queue,
                                            [self.edgar_url],
                                            [self.dir_work],
                                            [self.sub_filings]))
            pool = mp.Pool()
            for x in mplist:
                pool.apply_async(mpw.singledownload, x, callback=self.collect)
            pool.close()
            pool.join()

            # Check if errors occured and deal with these
            if len(self.errors) > 0:
                tqdm.write("Attempt %s: %s out of %s downloads failed."
                           % (attempt + 1,
                              len(self.errors),
                              len(self.url_list))
                           )
                tqdm.write("Attempt %s to download the filings "
                           "will start shortly."
                           % (attempt + 2))
                urls_queue = [item[0] for item in self.errors]
                self.errors = []
                attempt += 1
                time.sleep(5)
        # Handling unsuccesful attempts
        if len(self.results) < len(self.url_list) and attempt >= 2:
            tqdm.write("The final attempt was unsuccesful."
                       " %s out %s files could not be downloaded."
                       % (len(self.errors), len(self.url_list)))
        # Output if an error occured, but it could be fixed
        elif attempt > 0:
            tqdm.write("Attempt %s: Success. %s out %s files were downloaded."
                       % (attempt + 1, len(self.results), len(self.url_list)))
        # Create a pandas df with the filings' files information
        try:
            self.temp_files = pd.DataFrame(self.results,
                                           columns=["url",
                                                    "temp_fname",
                                                    "dt_accessed"])
            self.temp_files = self.temp_files.merge(self.index_slice,
                                                    left_on="url",
                                                    right_on="File Name")
        except Exception as e:
            tqdm.write(e)
        # Close the second (lower) progress bar
        if self.show_progress:
            self.bar.close()

    def splitfiles(self, text_only=True, **kwargs):
        """Extract the original documents from the temporary file from EDGAR.

        The archieved filings on EDGAR contain multiple files bundled in a
        .txt-file. This method splits the files into its original components.

        Args:
            text_only (bool): If True, only html and txt files are saved.
                All other (media) files are discarded. Defaults to True.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None

        Raises:
            None

        """
        result = pd.DataFrame(columns=["url",
                                       "seq",
                                       "server_fname",
                                       "type",
                                       "desc",
                                       "local_fname",
                                       ])
        # Iterate over all documents to split files
        for index, row in self.temp_files.iterrows():
            # Get get filename and make sure path exists
            tmp_fname = self.dir_work + self.sub_filings + row["temp_fname"]
            fname_part = create_filename(row, **kwargs)
            fname_full = self.dir_work + self.sub_filings + str(fname_part)
            fname_full = t.finduniquefname(fname_full, exact=False, **kwargs)
            path = os.path.dirname(fname_full)
            if os.path.isdir(path) is False:
                os.makedirs(path)

            # Open the temporary file and read the content
            f = open(tmp_fname)
            txt = f.read()
            f.close()

            # Extract the filing metadata, write to file & append to DataFrame
            m_sec_header = re.search("<SEC-HEADER>.*</SEC-HEADER>",
                                     txt,
                                     re.DOTALL)
            b_sec_header = m_sec_header.group(0)
            local_fname = fname_full + "header.txt"
            out = open(local_fname, "w")
            out.write(b_sec_header)
            out.close()
            result.loc[result.shape[0]] = ([row["url"],
                                            0,
                                            "HEADER",
                                            "SEC Header",
                                            "Header file",
                                            local_fname])

            # Extract all other document parts
            starts = ([a.end() for a in list(re.finditer("<DOCUMENT>", txt))])
            ends = ([a.start() for a in list(re.finditer("</DOCUMENT>", txt))])

            # For each documents path: Extract information, save to file and
            # append to dataframe; depending on text_only only save text or all
            # information
            for i in range(len(starts)):
                content = txt[starts[i]:ends[i]]
                text_startpos = content.find("<TEXT>\n") + 7
                text_endpos = content.find("</TEXT") - 1
                text = content[text_startpos:text_endpos]

                search = ("<TYPE>(?P<TYPE>.+?)[\\n]+?"
                          "<SEQUENCE>(?P<SEQ>.+?)[\\n]+"
                          "<FILENAME>(?P<FNAME>.+?)[\\n]"
                          "(<DESCRIPTION>(?P<DESC>.+?)[\\n])?")
                try:
                    info = re.search(search, content, re.MULTILINE).groupdict()
                except:
                    print(content)
                    break
                if info["TYPE"] != "GRAPHIC":
                    local_fname = fname_full + str(info["SEQ"]) + ".html"
                    out = open(local_fname, "w")
                    out.write(text)
                    out.close()
                else:
                    if text_only is True:
                        break
                    local_fname = fname_full + str(info["SEQ"]) + ".jpg"
                    f = io.BytesIO(bytes(text, encoding="ascii"))
                    uu.decode(f, local_fname, quiet=True)
                    f.close()

                result.loc[result.shape[0]] = ([row["url"],
                                               info["SEQ"],
                                               info["FNAME"],
                                               info["TYPE"],
                                               info["DESC"],
                                               local_fname])
        self.docs = result

    def replace_img(self):
        """Replace image path in documents with local filename.

        Args:
            None

        Returns:
            None

        Raises:
            None

        """
        for url in self.docs.url.unique():
            extract_full = self.docs.loc[self.docs.url == url]
            extract_graphic = extract_full.loc[extract_full.type == "GRAPHIC"]
            extract_other = extract_full.loc[extract_full.type != "GRAPHIC"]
            if extract_graphic.shape[0] > 0:
                for index, row in extract_other.iterrows():
                    f = open(row["local_fname"], "r")
                    txt = f.read()
                    local = extract_graphic.local_fname.values
                    server = extract_graphic.server_fname.values
                    for i in range(len(local)):
                        txt = txt.replace(server[i], local[i].split("/")[-1])
                    f.close()
                    f = open(row["local_fname"], "w")
                    f.write(txt)
                    f.close

    def delete_tempfiles(self):
        """Delete temporary raw downloaded files.

        Args:
            None

        Returns:
            None

        Raises:
            None

        """
        for index, row in self.temp_files.iterrows():
            fname = self.dir_work + self.sub_filings + row["temp_fname"]
            try:
                os.remove(fname)
            except:
                time.sleep(3)
                os.remove(fname)


def create_filename(row, fname_form="%Y%m%d_%company_", **kwargs):
    """Create filename for local EDGAR filing' files.

    Args:
         row (pandas series): Contains the columns "Date Filed",
            "Company Name" and "url".
         fname_form (str): String with the filename format.
            Possible parameters:
            * %org - orignal filename on server
            * %company - Company Name
            * All date variables based on strftime.
              See: http://strftime.org/
            Defaults to "%Y%m%d_%company_".
         **kwargs: Arbitrary keyword arguments.

    Returns:
         fname (str): The filename for the specific filing.

    Raises:
        None

    """
    # Extract the useful parts from the original filename on server
    splits = row["url"].split("/")
    org = splits[2] + "_" + splits[3].split(".")[0]

    # Extract unwanted characters from the company name
    company = (row["Company Name"]
               .replace(" ", "")
               .replace(",", "")
               .replace(".", "")
               .replace("/", "")
               .replace("\\", "")
               .lower()
               )
    # Replace the parameters with real values
    fname = fname_form.replace("%org", org).replace("%company", company)
    tdate = datetime.datetime.strptime(row["Date Filed"], "%Y-%m-%d")
    while fname.find("%") >= 0:
        char = "%" + fname[fname.find("%") + 1]
        fname = fname.replace(char, tdate.strftime(char))
    return fname
